package com.gxzc.zen.common.util

import com.gxzc.zen.common.dto.ZenMultipartFileDTO
import com.gxzc.zen.common.properties.UploadProperties
import org.slf4j.LoggerFactory
import java.io.*
import java.nio.file.Files
import java.nio.file.Paths
import java.util.concurrent.ConcurrentHashMap

/**
 * 上传 工具类
 * @author NorthLan
 * @date 2018/5/19
 * @url https://noahlan.com
 */
object UploadUtil {
    private val logger = LoggerFactory.getLogger(UploadUtil::class.java)

    private var uploadProperties = SpringContextHolder.getBean(UploadProperties::class.java)
        get() {
            if (field == null) {
                field = SpringContextHolder.getBean(UploadProperties::class.java)
            }
            return field
        }

    private val uploadedMap = ConcurrentHashMap<String, Int>() // 分片上传数

    /**
     * 正常上传<br>
     * 保留源 文件夹结构/文件名
     */
    fun upload(fileDTO: ZenMultipartFileDTO): Boolean {
        val tmpPath = uploadProperties!!.tmpPath!!
        val dataPath = uploadProperties!!.dataPath!!
        if (validateRequest(fileDTO)) {
            val chunkFilename = getChunkFilename(tmpPath, fileDTO.chunkNumber, fileDTO.identifier)
            val directory = File(tmpPath)
            if (!directory.exists()) {
                Files.createDirectories(Paths.get(tmpPath))
            }
            // 文件较小 直接transfer
            fileDTO.file!!.transferTo(File(chunkFilename))
            // 检查分块完整性
            if (checkChunks(fileDTO.identifier!!, fileDTO.totalChunks!!)) {
                // 合并
                mergeChunks(tmpPath, dataPath, fileDTO)
            }
            return true
        } else {
            return false
        }
    }

    fun uploadGet(fileDTO: ZenMultipartFileDTO): Boolean {
        val tmpPath = uploadProperties!!.tmpPath!!
        return if (validateRequest(fileDTO)) {
            val existsFile = File(getChunkFilename(tmpPath, fileDTO.chunkNumber, fileDTO.identifier))
            existsFile.exists()
        } else {
            false
        }
    }

    /**
     * 合并所有分块
     * 支持 文件夹合并
     */
    private fun mergeChunks(sourceRootPath: String, destRootPath: String, fileDTO: ZenMultipartFileDTO) {
        // 源文件的文件夹信息 然后组合拼接
        val filename = fileDTO.filename!!
        val relativePath = fileDTO.relativePath!!
        val totalChunks = fileDTO.totalChunks!!
        logger.debug("start merging chunks for [$filename]")

        val folderRelativePath = relativePath.replace(filename, "")
        // 示例 "sql/xx.sql" -> "sql/"
        // "xx.sql" -> ""

        val destDir = File("$destRootPath/$folderRelativePath")
        if (!destDir.exists()) {
            Files.createDirectories(Paths.get("$destRootPath/$folderRelativePath"))
        }
        // 目标文件流 通过文件名来拼接
        val destOutputStream = BufferedOutputStream(FileOutputStream("$destRootPath/$folderRelativePath$filename"))

        val buffer = ByteArray(1024)
        var readBytesLength: Int
        for (i: Int in 1..totalChunks) {
            val sourceFile = File(getChunkFilename(sourceRootPath, i, fileDTO.identifier))
            val sourceInputStream = BufferedInputStream(FileInputStream(sourceFile))

            readBytesLength = sourceInputStream.read(buffer)
            while (readBytesLength != -1) {
                destOutputStream.write(buffer, 0, readBytesLength)
                readBytesLength = sourceInputStream.read(buffer)
            }
            sourceInputStream.close()
            // 删除分片
            sourceFile.delete()
        }
        //
        destOutputStream.flush()
        destOutputStream.close()

        logger.debug("merging successful.")
    }

    /**
     * 检查分块是否上传完毕
     */
    private fun checkChunks(identifier: String, totalChunks: Int): Boolean {
        val cleanIdentifier = cleanIdentifier(identifier)!!
        var chunksNow = uploadedMap.getOrDefault(cleanIdentifier, 0)
        return if (totalChunks == ++chunksNow) {
            uploadedMap.remove(cleanIdentifier)
            true
        } else {
            uploadedMap[cleanIdentifier] = chunksNow
            false
        }
    }

    /**
     * 移除多余的identifier字符
     */
    private fun cleanIdentifier(identifier: String?): String? {
        return identifier?.replace(Regex("[^0-9A-Za-z_-]"), "")
    }

    /**
     * 获取 分片文件名
     */
    private fun getChunkFilename(path: String, chunkNumber: Int?, identifier: String?): String {
        return "$path/upload-${cleanIdentifier(identifier)}.$chunkNumber"
    }

    /**
     * validate request multipart chunks
     */
    private fun validateRequest(fileDTO: ZenMultipartFileDTO): Boolean {
        val identifier = cleanIdentifier(fileDTO.identifier)
        val chunkNumber = fileDTO.chunkNumber
        val chunkSize = fileDTO.chunkSize
        val totalSize = fileDTO.totalSize
        val filename = fileDTO.filename

        if (chunkNumber == null || chunkNumber <= 0 ||
                chunkSize == null || chunkSize <= 0 ||
                totalSize == null || totalSize <= 0 ||
                identifier == null || identifier.isEmpty() ||
                filename == null || filename.isEmpty()) {
            return false
        }

        val numberOfChunks = Math.max(Math.floor(totalSize / (chunkSize * 1.0)), 1.0).toInt()
        if (chunkNumber > numberOfChunks) {
            return false
        }

        // is the file too large?
        uploadProperties?.let {
            val maxFileSize = it.maxFileSize
            if (maxFileSize != null && maxFileSize > 0) {
                if (totalSize > maxFileSize) {
                    logger.error("filesize limit: [${maxFileSize / 1024 / 1024} MB], now [${totalSize / 1024 / 1024} MB]")
                    return false
                }
            }
        }


        val file = fileDTO.file
        if (file != null) {
            // The chunk in the POST request isn't the correct size
            if (chunkNumber < numberOfChunks && file.size != chunkSize) {
                logger.error("The chunk in the POST request isn't the correct size")
                return false
            }
            // The chunks in the POST is the last one, and the fil is not the correct size
            if (numberOfChunks > 1 && chunkNumber == numberOfChunks && file.size != (totalSize % chunkSize) + chunkSize) {
                logger.error("The chunks in the POST is the last one, and the fil is not the correct size")
                return false
            }
            // The file is only a single chunk, and the data size does not fit
            if (numberOfChunks == 1 && file.size != totalSize) {
                logger.error("The file is only a single chunk, and the data size does not fit")
                return false
            }
        }
        return true
    }
}